#include <stdlib.h>
#include <string.h>

#include "block.h"
#include "area.h"
#include "file_object.h"

typedef struct _FileData *FileData;

struct file_object {
  file_object_type_t type;
  unsigned int flags;
  int reference;
  FileData data;
};

static FileData FileObject_GetData(FileObject object);
static FileData FileObject_SetData(FileObject object, FileData data);
static FileData FileObject_DataCreate(FileObject object);
static FileData FileObject_DataDestroy(FileObject object);

/******** for FileObject operation ********/

BlockPool FileObject_GetPool(FileObject object)
{
  return Area_GetPool((Area)object);
}

static int FileObject_Read(FileObject object, struct file_object *info)
{
  return Area_Read((Area)object, 0, sizeof(*info), info);
}

static int FileObject_Write(FileObject object, struct file_object *info)
{
  return Area_Write((Area)object, 0, sizeof(*info), info);
}

file_object_type_t FileObject_GetType(FileObject object)
{
  struct file_object info;
  FileObject_Read(object, &info);
  return info.type;
}

file_object_type_t FileObject_SetType(FileObject object, file_object_type_t type)
{
  struct file_object info;
  FileObject_Read(object, &info);
  info.type = type;
  FileObject_Write(object, &info);
  return info.type;
}

unsigned int FileObject_GetFlags(FileObject object)
{
  struct file_object info;
  FileObject_Read(object, &info);
  return info.flags;
}

unsigned int FileObject_SetFlags(FileObject object, unsigned int flags)
{
  struct file_object info;
  FileObject_Read(object, &info);
  info.flags = flags;
  FileObject_Write(object, &info);
  return info.flags;
}

int FileObject_GetReference(FileObject object)
{
  struct file_object info;
  FileObject_Read(object, &info);
  return info.reference;
}

int FileObject_SetReference(FileObject object, int reference)
{
  struct file_object info;
  FileObject_Read(object, &info);
  info.reference = reference;
  FileObject_Write(object, &info);
  return info.reference;
}

int FileObject_ClearReference(FileObject object)
{
  return FileObject_SetReference(object, 0);
}

int FileObject_AddReference(FileObject object)
{
  return FileObject_SetReference(object, FileObject_GetReference(object) + 1);
}

int FileObject_DelReference(FileObject object)
{
  return FileObject_SetReference(object, FileObject_GetReference(object) - 1);
}

static FileData FileObject_GetData(FileObject object)
{
  struct file_object info;
  FileObject_Read(object, &info);
  return info.data;
}

static FileData FileObject_SetData(FileObject object, FileData data)
{
  struct file_object info;
  FileObject_Read(object, &info);
  info.data = data;
  FileObject_Write(object, &info);
  return info.data;
}

FileObject FileObject_Create(BlockPool pool, file_object_type_t type)
{
  FileObject object;
  struct file_object info;

  object = (FileObject)Area_Create(pool);
  Area_Resize((Area)object, sizeof(info));

  memset(&info, 0, sizeof(info));
  info.type      = type;
  info.flags     = 0;
  info.reference = 0;
  info.data      = NULL;
  FileObject_Write(object, &info);

  FileObject_DataCreate(object);

  return object;
}

FileObject FileObject_Destroy(FileObject object)
{
  FileObject_DataDestroy(object);
  return (FileObject)Area_Destroy((Area)object);
}

/******** for FileData operation ********/

int FileObject_DataGetSize(FileObject object)
{ return Area_GetSize((Area)FileObject_GetData(object)); }

int FileObject_DataRead(FileObject object, int offset, int size, void *buffer)
{ return Area_Read((Area)FileObject_GetData(object), offset, size, buffer); }

int FileObject_DataWrite(FileObject object, int offset, int size, const void *buffer)
{ return Area_Write((Area)FileObject_GetData(object), offset, size, buffer); }

int FileObject_DataCompare(FileObject object, int offset, int size, const void *buffer)
{ return Area_Compare((Area)FileObject_GetData(object), offset, size, buffer); }

int FileObject_DataClear(FileObject object)
{ return Area_Clear((Area)FileObject_GetData(object)); }

int FileObject_DataExtend(FileObject object, int size)
{ return Area_Extend((Area)FileObject_GetData(object), size); }

int FileObject_DataReduce(FileObject object, int size)
{ return Area_Reduce((Area)FileObject_GetData(object), size); }

int FileObject_DataResize(FileObject object, int size)
{ return Area_Resize((Area)FileObject_GetData(object), size); }

static FileData FileObject_DataCreate(FileObject object)
{
  BlockPool pool = FileObject_GetPool(object);
  return FileObject_SetData(object, (FileData)Area_Create(pool));
}

static FileData FileObject_DataDestroy(FileObject object)
{
  return FileObject_SetData(object, (FileData)Area_Destroy((Area)FileObject_GetData(object)));
}
